#include "relations.h"

void pollardsRho(mpz_t n, mpz_t res) {
    mpz_t x, y, d, c, t1;
    //mpz_inits(x, y, d, c, t1, 0);
    uint32_t k = 0;
    
    mpz_init(x);
    mpz_init(y);
    mpz_init(d);
    mpz_init(c);
    mpz_init(t1);

    for (mpz_urandomm(c, randstate, n), mpz_urandomm(x, randstate, n), mpz_set(y, x), mpz_set_ui(d, 1);;) {
        for (;;) {
            mpz_mul(x, x, x); mpz_add(x, x, c); mpz_mod(x, x, n);
            
            mpz_mul(y, y, y); mpz_add(y, y, c); mpz_mod(y, y, n);
            mpz_mul(y, y, y); mpz_add(y, y, c); mpz_mod(y, y, n);
            
            mpz_sub(t1, x, y); mpz_abs(t1, t1);
            mpz_mul(d, d, t1); mpz_mod(d, d, n);        
            if (((++k) & 1023) == 0) {
                mpz_gcd(t1, n, d);
                mpz_set(d, t1);

                if (mpz_cmp_ui(d, 1) == 0) continue;
                if (mpz_cmp(d, n) == 0) break;
                
                mpz_set(res, d);
                mpz_divexact(n, n, d);
                goto exit;
                //mpz_clears(x, y, d, c, t1, 0);       
                //return;
            }
        }
    }
exit:    
//    mpz_clears(x, y, d, c, t1, 0);
    mpz_clear(x);
    mpz_clear(y);
    mpz_clear(d);
    mpz_clear(c);
    mpz_clear(t1);    
}

void pollard_rho(mpz_t n, uint64_t a, uint64_t p, mpz_t res){
	mpz_t x, x1, y, P;
	mpz_t t1, t2;
	uint64_t k, l, i;
	
	//mpz_inits (t1, t2, NULL);
    mpz_init(t1);
    mpz_init(t2);
    
	mpz_init_set_si (y, 2);
	mpz_init_set_si (x, 2);
	mpz_init_set_si (x1, 2);
	mpz_init_set_ui (P, 1);
	k = 1;
	l = 1;
	
	while (mpz_cmp_ui (n, 1) != 0)
    {
		for (;;)
		{
			do
			{
				if (p != 0)
				{
					mpz_powm_ui (x, x, p, n);
					mpz_add_ui (x, x, a);
				}
				else
				{
					mpz_mul (t1, x, x);
					mpz_mod (x, t1, n);
					mpz_add_ui (x, x, a);
				}
				
				mpz_sub (t1, x1, x);
				mpz_mul (t2, P, t1);
				mpz_mod (P, t2, n);
				
				if (k % 32 == 1)
				{
					mpz_gcd (t1, P, n);
					if (mpz_cmp_ui (t1, 1) != 0)
						goto factor_found;
					mpz_set (y, x);
				}
			}
			while (--k != 0);
			
			mpz_gcd (t1, P, n);
			if (mpz_cmp_ui (t1, 1) != 0)
				goto factor_found;
			
			mpz_set (x1, x);
			k = l;
			l = 2 * l;
			for (i = 0; i < k; i++)
			{
				if (p != 0)
				{
					mpz_powm_ui (x, x, p, n);
					mpz_add_ui (x, x, a);
				}
				else
				{
					mpz_mul (t1, x, x);
					mpz_mod (x, t1, n);
					mpz_add_ui (x, x, a);
				}
			}
			mpz_set (y, x);
		}
		
    factor_found:
		do
		{
			if (p != 0)
			{
				mpz_powm_ui (y, y, p, n); mpz_add_ui (y, y, a);
			}
			else
			{
				mpz_mul (t1, y, y);
				mpz_mod (y, t1, n);
				mpz_add_ui (y, y, a);
			}
			mpz_sub (t1, x1, y);
			mpz_gcd (t1, t1, n);
		}
		while (mpz_cmp_ui (t1, 1) == 0);
		
		mpz_divexact (n, n, t1);	/* divide by t1, before t1 is overwritten */
		mpz_set(res, t1);
		break;
		/*
		 mpz_out_str (stdout, 10, t1);
		 fflush (stdout);
		 fputc (' ', stdout);
		 
		 mpz_mod (x, x, n);
		 mpz_mod (x1, x1, n);
		 mpz_mod (y, y, n);
		 if (mpz_probab_prime_p (n, 10))
		 {
		 mpz_out_str (stdout, 10, n);
		 fflush (stdout);
		 fputc (' ', stdout);
		 break;
		 }*/
    }
	
	//mpz_clears (P, t2, t1, x1, x, y, NULL);
    mpz_clear(P);
    mpz_clear(t2);
    mpz_clear(t1);
    mpz_clear(x1);
    mpz_clear(x);
    mpz_clear(y);
    
}

uint64_t bgcd(uint64_t a, uint64_t b) {
	uint64_t pof2 = 0;
	
	if(!a || !b) return a | b;
	while(!((a | b) & 1)){// a & b are even
		a >>= 1; b >>= 1; ++pof2;
	}
	
	while(!(a & 1)) a >>= 1;
	while(!(b & 1)) b >>= 1;
	while(a > 1 && b > 1 && a != b){
		if(a > b){
			a = (a - b) >> 1;
			while(!(a & 1)) a >>= 1;
		}
		else{
			b = (b - a) >> 1;
			while(!(b & 1)) b >>= 1;
		}
	}
	return (a == b ? a : 1) << pof2;
}

int coprime(int64_t x, int64_t y) {
    if ((x | y) & 1LL) { // even
        if (y % 3 == 0 && x % 3 == 0) return 0;
        if (x < 0) x = -x;        
        if (bgcd(x, y) > 1) return 0; // y > 0
        return 1;
    }
    return 0;
}

int binSearch(sieveBase *B, int64_t x, int64_t y, int64_t p) {
    int leftBound = 0, rightBound = B->prBaseSize, midPoint;
    while (leftBound + 1 < rightBound) {
        midPoint = (leftBound + rightBound) >> 1;
        if (B->prBase[midPoint].p < p) {
            leftBound = midPoint; continue;
        }
        else if (B->prBase[midPoint].p > p) {
            rightBound = midPoint; continue;
        }
        
        while (B->prBase[--midPoint].p == p);
        ++midPoint;
        
        while (B->prBase[midPoint].p == p) {
            if ((x - y * B->prBase[midPoint].r) % p == 0) return midPoint;
            ++midPoint;
        }
        break;
    }
    return -1;
}

#ifndef USEPRINDEX
int binSearchBigPrime(sieveBase *B, int64_t x, int64_t y, int16_t delta, mpz_t val, factorization *result) {
    int leftBound = 0, rightBound = B->prBaseSize, midPoint;
    while (leftBound + 1 < rightBound) {
        midPoint = (leftBound + rightBound) >> 1;

        if (B->prBase[midPoint].delta < delta) {
            leftBound = midPoint; continue;
        }
        else if (B->prBase[midPoint].delta > delta) {
            rightBound = midPoint; continue;
        }
        
        while (--midPoint >= 0 && B->prBase[midPoint].delta == delta);
        
        while (++midPoint < B->prBaseSize && B->prBase[midPoint].delta == delta) {
            if ((x - y * B->prBase[midPoint].r) % B->prBase[midPoint].p == 0) {

                if (!mpz_divisible_ui_p(val, B->prBase[midPoint].p)) continue;

                result->alphas[result->alphasNumber][0] = midPoint;
                result->alphas[result->alphasNumber][1] = 0;
                do {
                    mpz_divexact_ui(val, val, B->prBase[midPoint].p);
                    result->alphas[result->alphasNumber][1]++;
                } while (mpz_divisible_ui_p(val, B->prBase[midPoint].p));
                result->alphasNumber++;                
            }
        }
        break;
    }
}
#endif

void checkXY(sieveBase *B, poly *P, int64_t x, int64_t y, factorization *result) {   
    int i, imax;
    mpz_t val;
#ifdef USEPRINDEX
    int64_t curp, curpIndex;
#endif

    CHECKTIME( clock_t ctime = clock(); )
    
    calcPoly(P, x, y, &val);
    mpz_abs(val, val);
    
    result->alphasNumber = 0;

    // special base
    for (i = 0; i < B->spBaseSize; ++i) {
        if (y % B->spEntries[i].p == 0) {
            result->alphas[result->alphasNumber][0] = -1 - i;
            result->alphas[result->alphasNumber][1] = 0;
            do {
                mpz_divexact_ui(val, val, B->spEntries[i].p);
                result->alphas[result->alphasNumber][1]++;
            } while (mpz_divisible_ui_p(val, B->spEntries[i].p));
            result->alphasNumber++;
        }
    }
    
    // already factored
    if (mpz_cmp_ui(val, 1) == 0) {
        result->bigPrimesNumber = 0;
        CHECKTIME( checkTime[0] += clock() - ctime; )
        return;
    }
    else if (mpz_probab_prime_p(val, 5)) {
        if (mpz_cmp_ui(val, B->B) <= 0) {
            result->alphas[result->alphasNumber][0] = binSearch(B, x, y, mpz_get_si(val));
            result->alphasNumber++;
            result->bigPrimesNumber = 0;
            CHECKTIME( checkTime[0] += clock() - ctime; )
            return;
        }
        
        if (mpz_cmp_ui(val, B->bigPrime) > 0) {
            result->bigPrimesNumber = -1;
            CHECKTIME( checkTime[0] += clock() - ctime; )
            return;
        }
        
        mpz_set(result->bp1, val);
        result->bigPrimesNumber = 1;
        CHECKTIME( checkTime[0] += clock() - ctime; )
        return;
    }

    CHECKTIME( checkTime[0] += clock() - ctime; ctime = clock(); )
    
    // factorization
    for (i = 0, imax = B->prBaseFirstMediumIndex; i < imax; ++i) {
        if ((y * B->prBase[i].r - x) % B->prBase[i].p == 0) {
            result->alphas[result->alphasNumber][0] = i;
            result->alphas[result->alphasNumber][1] = 0;
            do {
                mpz_divexact_ui(val, val, B->prBase[i].p);
                result->alphas[result->alphasNumber][1]++;
            } while (mpz_divisible_ui_p(val, B->prBase[i].p));
            result->alphasNumber++;
            
            if (mpz_cmp_ui(val, 1) == 0) {
                result->bigPrimesNumber = 0;
                CHECKTIME( checkTime[1] += clock() - ctime; )
                return;
            }
            else if (mpz_probab_prime_p(val, 5)) {
                if (mpz_cmp_ui(val, B->B) <= 0) {
                    result->alphas[result->alphasNumber][0] = binSearch(B, x, y, mpz_get_si(val));
                    result->alphasNumber++;
                    result->bigPrimesNumber = 0;
                    CHECKTIME( checkTime[1] += clock() - ctime; )
                    return;
                }
                
                if (mpz_cmp_ui(val, B->bigPrime) > 0) {
                    result->bigPrimesNumber = -1;
                    CHECKTIME( checkTime[1] += clock() - ctime; )
                    return;
                }
                
                mpz_set(result->bp1, val);
                result->bigPrimesNumber = 1;
                CHECKTIME( checkTime[1] += clock() - ctime; )
                return;
            }
        }
    }
    
    CHECKTIME( checkTime[1] += clock() - ctime; ctime = clock(); )
        
    // medium base <- traces
    int16_t tshift = x & (BLOCK_SIZE - 1);    
    for (i = 0, imax = B->traceSize; i < imax; ++i) {
        if (B->traces[i].shift == tshift) {
#ifdef USEPRINDEX
            curpIndex = B->traces[i].prIndex;
            curp = B->prBase[curpIndex].p;

            if (!mpz_divisible_ui_p(val, curp)) continue;
            result->alphas[result->alphasNumber][0] = curpIndex;
            result->alphas[result->alphasNumber][1] = 0;
            do {
                mpz_divexact_ui(val, val, curp);
                result->alphas[result->alphasNumber][1]++;
            } while (mpz_divisible_ui_p(val, curp));
            result->alphasNumber++;
#else
            binSearchBigPrime(B, x, y, B->traces[i].delta, val, result);
#endif            

            // we could also check if result->alphasNumber > tAlphasNumber
            if (mpz_cmp_ui(val, 1) == 0) {
                result->bigPrimesNumber = 0;
                CHECKTIME( checkTime[2] += clock() - ctime; )
                return;
            }
            else if (mpz_probab_prime_p(val, 5)) {
                if (mpz_cmp_ui(val, B->B) <= 0) {
                    result->alphas[result->alphasNumber][0] = binSearch(B, x, y, mpz_get_si(val));
                    result->alphasNumber++;
                    result->bigPrimesNumber = 0;
                    CHECKTIME( checkTime[2] += clock() - ctime; )
                    return;
                }
                
                if (mpz_cmp_ui(val, B->bigPrime) > 0) {
                    result->bigPrimesNumber = -1;
                    CHECKTIME( checkTime[2] += clock() - ctime; )
                    return;
                }
                
                mpz_set(result->bp1, val);
                result->bigPrimesNumber = 1;
                CHECKTIME( checkTime[2] += clock() - ctime; )
                return;
            }
        }
    }
    
    CHECKTIME( checkTime[2] += clock() - ctime; ctime = clock(); )
    
    // big chunks
    chunk *curChunk = B->bigChunks + B->currentChunkIndex - 1;
    for (i = 0, imax = curChunk->size; i < imax; ++i) {
        if (curChunk->entries[i].shift == tshift) {
#ifdef USEPRINDEX
            curpIndex = curChunk->entries[i].prIndex;
            curp = B->prBase[curpIndex].p;
            
            if (!mpz_divisible_ui_p(val, curp)) continue;
            result->alphas[result->alphasNumber][0] = curpIndex;
            result->alphas[result->alphasNumber][1] = 0;
            do {
                mpz_divexact_ui(val, val, curp);
                result->alphas[result->alphasNumber][1]++;
            } while (mpz_divisible_ui_p(val, B->prBase[i].p));
            result->alphasNumber++;
#else            
            binSearchBigPrime(B, x, y, curChunk->entries[i].delta, val, result);
#endif            
            
            // we could also check if result->alphasNumber > tAlphasNumber
            if (mpz_cmp_ui(val, 1) == 0) {
                result->bigPrimesNumber = 0;
                CHECKTIME( checkTime[3] += clock() - ctime; )
                return;
            }
            else if (mpz_probab_prime_p(val, 5)) {
                if (mpz_cmp_ui(val, B->B) <= 0) {
                    result->alphas[result->alphasNumber][0] = binSearch(B, x, y, mpz_get_si(val));
                    result->alphasNumber++;
                    result->bigPrimesNumber = 0;
                    CHECKTIME( checkTime[3] += clock() - ctime; )
                    return;
                }
                
                if (mpz_cmp_ui(val, B->bigPrime) > 0) {
                    result->bigPrimesNumber = -1;
                    CHECKTIME( checkTime[3] += clock() - ctime; )
                    return;
                }
                
                mpz_set(result->bp1, val);
                result->bigPrimesNumber = 1;
                CHECKTIME( checkTime[3] += clock() - ctime; )
                return;
            }
        }
    }
    
    // 2BP?
    mpz_t temp; mpz_init(temp);
    mpz_set_ui(temp, B->bigPrime);
    mpz_mul(temp, temp, temp);
    if (mpz_cmp(val, temp) > 0) {
        result->bigPrimesNumber = -1;
        CHECKTIME( checkTime[3] += clock() - ctime; )
        return;
    }
    // redundantly
    //mpz_set_ui(temp, B->B);
    //mpz_mul(temp, temp, temp);
    //if (mpz_cmp(val, temp) < 0) {
    //    result->bigPrimesNumber = -1;
    //    return;
    //}
    
    CHECKTIME( checkTime[3] += clock() - ctime; ctime = clock(); )
       
    pollard_rho(val, 1L, 0, temp);
    if (mpz_cmp_ui(val, B->bigPrime) > 0 || mpz_cmp_ui(temp, B->bigPrime) > 0) {
        result->bigPrimesNumber = -1;
        CHECKTIME( checkTime[4] += clock() - ctime; )
        return;
    }
    
    result->bigPrimesNumber = 2;
    mpz_set(result->bp1, val);
    mpz_set(result->bp2, temp);
    
    //mpz_clears(val, temp, 0);
    mpz_clear(val);
    mpz_clear(temp);
    
    CHECKTIME( checkTime[4] += clock() - ctime; )
}

void serialize(FILE *fout, int64_t x, int64_t y, factorization *facF, factorization *facG) {    
    fprintf(fout, "%lld,%lld: %d%d ", x, y, facF->bigPrimesNumber, facG->bigPrimesNumber);
    switch(facF->bigPrimesNumber) {
        case 0:
            fprintf(fout, "1");
            break;
        case 1:
            gmp_fprintf(fout, "%Zd", facF->bp1);
            break;
        default:
            gmp_fprintf(fout, "%Zdx%Zd", facF->bp1, facF->bp2);
            break;
    }
    fprintf(fout, ", ");
    switch(facG->bigPrimesNumber) {
        case 0:
            fprintf(fout, "1");
            break;
        case 1:
            gmp_fprintf(fout, "%Zd", facG->bp1);
            break;
        default:
            gmp_fprintf(fout, "%Zdx%Zd", facG->bp1, facG->bp2);
            break;
    }
    fprintf(fout, "; [%d %d]\n", facF->alphasNumber, facG->alphasNumber);
    
    int i;
    for (i = 0; i < facF->alphasNumber; ++i) {
        fprintf(fout, "%d %d\n", facF->alphas[i][0], facF->alphas[i][1]);
    }
    for (i = 0; i < facG->alphasNumber; ++i) {
        fprintf(fout, "%d %d\n", facG->alphas[i][0], facG->alphas[i][1]);
    }
}
